from copy import deepcopy
from sklearn.base import clone
from sklearn.model_selection import GridSearchCV, StratifiedKFold
from sklearn.neighbors import KNeighborsClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.linear_model import LogisticRegression
from xgboost import XGBClassifier
from imblearn.pipeline import Pipeline
from imblearn.over_sampling import SMOTE

# Validación cruzada estratificada con 5 folds
CV = StratifiedKFold(n_splits=5, shuffle=True, random_state=42)

# Diccionario de métodos disponibles
MODELS = {
    "knn":                KNeighborsClassifier(),
    "random_forest":      RandomForestClassifier(random_state=42),
    "decision_tree":      DecisionTreeClassifier(random_state=42),
    "logistic_regression":LogisticRegression(max_iter=1000, random_state=42),
    "xgboost":            XGBClassifier(eval_metric="logloss", random_state=42)
}

# Hiperparámetros para GridSearchCV
PARAMS = {
    "knn": {
        "n_neighbors": [3, 5, 7, 9],
        "weights": ["uniform", "distance"],
        "metric": ["euclidean", "manhattan"]
    },
    "random_forest": {
        "n_estimators": [100, 200],
        "max_depth": [None, 10, 20],
        "min_samples_split": [2, 5]
    },
    "decision_tree": {
        "max_depth": [None, 10, 20],
        "min_samples_split": [2, 5],
        "criterion": ["gini", "entropy", "log_loss"]
    },
    "logistic_regression": {
        "C": [0.01, 0.1, 1, 10, 100],
        "solver": ["liblinear", "saga"],
        "penalty": ["l1", "l2"],
    },
    "xgboost": {
        "n_estimators": [10, 50, 100, 150, 500],
        "max_depth": [2, 3, 5, 7, 10],
        "learning_rate": [0.01, 0.1, 0.2]
    }
}

def _grid_with_smote(estimator, params):
    """
    GridSearchCV aplicando SMOTE
    """
    pipe = Pipeline([
        ("smote", SMOTE(random_state=42)),
        ("clf", estimator)
    ])
    params = {f"clf__{k}": v for k, v in params.items()}
    return GridSearchCV(pipe, params, scoring="f1",
                        cv=CV, verbose=0)

def grid_search(estimator, params):
    """
    GridSearchCV sin técnicas de balanceo
    """
    return GridSearchCV(estimator, params, scoring='f1', cv=CV, verbose=0)

def _make(name, *, variant="normal", pos_weight=None):
    """
    GridSearchCV (smote o normal) con el método correspondiente (con o sin cost-sensitive)
    """
    est = clone(MODELS[name])
    grid = deepcopy(PARAMS[name])

    # cost-sensitive
    if variant.startswith("cs"):
        if name == "xgboost":
            est.set_params(scale_pos_weight=pos_weight)
        else:
            est.set_params(class_weight="balanced")

    # GridSearchCV
    if "smote" in variant:
        return _grid_with_smote(est, grid)
    else:
        return grid_search(est, grid)

# Modelos normales (sin balanceo)
def knn(): return _make("knn")
def random_forest(): return _make("random_forest")
def decision_tree(): return _make("decision_tree")
def logistic_regression(): return _make("logistic_regression")
def xgboost(): return _make("xgboost")

# Modelos con SMOTE
def knn_smote():return _make("knn", variant="smote")
def random_forest_smote(): return _make("random_forest", variant="smote")
def decision_tree_smote(): return _make("decision_tree", variant="smote")
def logistic_regression_smote(): return _make("logistic_regression", variant="smote")
def xgboost_smote(): return _make("xgboost", variant="smote")

# Modelos cost-sensitive
def random_forest_cs(): return _make("random_forest", variant="cs")
def decision_tree_cs(): return _make("decision_tree", variant="cs")
def logistic_regression_cs(): return _make("logistic_regression", variant="cs")
def xgboost_cs(pos_w): return _make("xgboost", variant="cs", pos_weight=pos_w)