import numpy as np
import pandas as pd
from imblearn.over_sampling import SMOTE
from sklearn.cluster import KMeans
from sklearn.metrics import pairwise_distances_argmin_min
from sklearn.preprocessing import StandardScaler

def split_holdout(df: pd.DataFrame, alzheimer_id="C0002395", esquizofrenia_id="C0036341"):
    """
        Divide el dataset en entrenamiento y test para dos enfermedades concretas (Alzhéimer y Esquizofrenia).
    """
    # Separar conjuntos
    train = df[(df.disease_id != alzheimer_id) & (df.disease_id != esquizofrenia_id)]
    test_alzheimer = df[df.disease_id == alzheimer_id]
    test_esquizofrenia = df[df.disease_id == esquizofrenia_id]

    # Datos de entrenamiento (sin las dos enfermedades)
    X_train = train.drop(columns=["disease_id", "drug_id", "class"])
    y_train = train["class"]

    # Alzheimer (conjunto de test)
    df_test_alz = test_alzheimer[["disease_id", "drug_id", "class"]].reset_index(drop=True)
    X_test_alz = test_alzheimer.drop(columns=["disease_id", "drug_id", "class"])
    y_test_alz = test_alzheimer["class"]

    # Esquizofrenia (conjunto de test)
    df_test_esq = test_esquizofrenia[["disease_id", "drug_id", "class"]].reset_index(drop=True)
    X_test_esq = test_esquizofrenia.drop(columns=["disease_id", "drug_id", "class"])
    y_test_esq = test_esquizofrenia["class"]

    return X_train, y_train, X_test_alz, y_test_alz, df_test_alz, X_test_esq, y_test_esq, df_test_esq

def scale(X_train, X_test):
    """
    Aplica escalado estándar (media=0, var=1) a los datos de entrenamiento y test.
    """
    scaler = StandardScaler()
    X_train_scaled = pd.DataFrame(scaler.fit_transform(X_train), columns=X_train.columns, index=X_train.index)
    X_test_scaled = pd.DataFrame(scaler.transform(X_test), columns=X_test.columns, index=X_test.index)
    return X_train_scaled, X_test_scaled

def oversample(X, y):
    """
    Aplica SMOTE para balancear clases en los datos de entrenamiento.
    """
    smote = SMOTE(random_state=42)
    return smote.fit_resample(X, y)

def cluster_downsample(X: pd.DataFrame, y: pd.Series):
    """
    Downsampling sobre la clase mayoritaria usando KMeans.
    """
    # Asegurar que X es DataFrame y y es Series
    if isinstance(X, np.ndarray):
        X = pd.DataFrame(X)
    if isinstance(y, (np.ndarray, list)):
        y = pd.Series(y)

    X = X.reset_index(drop=True)
    y = y.reset_index(drop=True)

    X0, y0 = X[y == 0], y[y == 0]
    X1, y1 = X[y == 1], y[y == 1]

    kmeans = KMeans(n_clusters=len(y1), random_state=42).fit(X0)
    centroids = kmeans.cluster_centers_
    reps, _ = pairwise_distances_argmin_min(centroids, X0)

    X0_red = X0.iloc[reps]
    y0_red = y0.iloc[reps]

    X_downsampling = pd.concat([X0_red, X1])
    y_downsampling = pd.concat([y0_red, y1])
    return X_downsampling, y_downsampling
