import importlib
import os
import pathlib
import pickle
from tqdm import tqdm
import matplotlib
import pandas as pd
from sklearn.model_selection import StratifiedKFold
from .metrics import recall_precision_at_k, summary_metrics, clasif_report, find_best_threshold
from . import models
matplotlib.use("Agg")

# Paths globales
ROOT = pathlib.Path(__file__).resolve().parents[1]
DATA = ROOT / "data"/ "unified"
RESULTS = ROOT / "results"
FIGURES = ROOT / "figures"

def _nested_cv_train(X, y, model_fn):
    """
    Ejecuta validación cruzada anidada:
    - Inner loop: GridSearchCV para buscar hiperparámetros
    - Outer loop: evalúa el rendimiento del modelo ajustado (con los mejores hiperparámetros del inner loop) en múltiples folds
    Devuelve: mejor modelo, hiperparámetros óptimos, umbral y métricas agregadas
    """
    if not isinstance(X, pd.DataFrame):X = pd.DataFrame(X)
    if not isinstance(y, pd.Series):y = pd.Series(y)

    outer_cv = StratifiedKFold(n_splits=5, shuffle=True, random_state=42)
    metrics_per_fold = []
    thresholds = []
    models = []
    best_params_fold = []
    for fold_idx, (train_idx, val_idx) in enumerate(
            tqdm(outer_cv.split(X, y), total=outer_cv.get_n_splits(X, y), desc="CV folds", position=0, leave=False)
    ):
        X_train, X_val = X.iloc[train_idx], X.iloc[val_idx]
        y_train, y_val = y.iloc[train_idx], y.iloc[val_idx]

        # Inner loop: GridSearchCV para encontrar los mejores hiperparámetros
        grid = model_fn()
        grid.fit(X_train, y_train)
        best_model = grid.best_estimator_
        best_params = grid.best_params_
        y_val_proba = best_model.predict_proba(X_val)[:, 1]

        # Mejor umbral
        thr, _ = find_best_threshold(y_val, y_val_proba)
        y_val_pred = (y_val_proba >= thr).astype(int)
        metrics_fold = summary_metrics(y_val, y_val_pred, y_val_proba)

        # Guardar resultados
        metrics_per_fold.append(metrics_fold)
        thresholds.append(thr)
        models.append(best_model)
        best_params_fold.append(best_params)

    # DataFrame con resultados de cada fold
    df_metrics_cv = pd.DataFrame(metrics_per_fold)
    # Elegir el mejor fold por f1_class1
    best_fold = df_metrics_cv["f1_class1"].idxmax()

    metrics_cv = {}
    for k in df_metrics_cv.columns:
        metrics_cv[f"{k}_cv_mean"] = df_metrics_cv[k].mean()
        metrics_cv[f"{k}_cv_std"] = df_metrics_cv[k].std()

    best_model = models[best_fold]
    best_params = best_params_fold[best_fold]
    best_thr = thresholds[best_fold]

    return best_model, best_params, best_thr, metrics_cv

def _evaluate_test_set(model, threshold, X_test, y_test, df_test, tag):
    """
    Evalúa el modelo entrenado sobre los conjuntos de test.
    Devuelve predicciones, métricas y curva precision-recall@K
    """
    y_proba = model.predict_proba(X_test)[:, 1]
    y_pred = (y_proba >= threshold).astype(int)

    df_preds = df_test.copy()
    df_preds["y_pred"] = y_pred
    df_preds["y_proba"] = y_proba

    df_k = recall_precision_at_k(y_test, y_proba, name_model=tag)
    summ = summary_metrics(y_test, y_pred, y_proba)
    report = clasif_report(y_test, y_pred)

    return df_k, summ, report, df_preds

def run_pipeline(df, model_name:str, pipeline_name:str):
    """
    Ejecuta todo el pipeline de clasificación:
    - Prepara los datos
    - Entrena con validación cruzada anidada
    - Evalúa en los conjuntos de test (Alzhéimer y Esquizofrenia)
    - Devuelve las predicciones, .pickle y métricas
    """
    # Preparar datos según el pipeline
    prep = importlib.import_module(f".pipelines.{pipeline_name}", package="code")
    X_train, y_train, X_test_alz, y_test_alz, df_test_alz, X_test_esq, y_test_esq, df_test_esq = prep.prepare(df, model_name)

    # Funcion del modelo a usar
    if pipeline_name == "oversampling_gridsearch":
        model_fn = getattr(models, f"{model_name}_smote")
    elif pipeline_name == "cost_sensitive":
        if model_name == "xgboost":
            pos_w = (y_train == 0).sum() / (y_train == 1).sum()
            model_fn = lambda: models.xgboost_cs(pos_w)
        else:
            model_fn = getattr(models, f"{model_name}_cs")
    else:
        model_fn = getattr(models, model_name)

    tag = f"{pipeline_name}-{model_name}"

    # Entrenamiento
    best_model, best_params, best_thr, metrics_cv = _nested_cv_train(X_train, y_train, model_fn)
    # Evaluar Alzheimer
    df_k_alz, summ_alz, report_alz, df_preds_alz = _evaluate_test_set(
        best_model, best_thr, X_test_alz, y_test_alz, df_test_alz, tag + "-alz"
    )
    # Evaluar Esquizofrenia
    df_k_esq, summ_esq, report_esq, df_preds_esq = _evaluate_test_set(
        best_model, best_thr, X_test_esq, y_test_esq, df_test_esq, tag + "-esq"
    )

    return {
        "alz": (df_k_alz, summ_alz, report_alz, df_preds_alz),
        "esq": (df_k_esq, summ_esq, report_esq, df_preds_esq),
        "cv_metrics": metrics_cv,
        "best_model": best_model,
        "params": best_params,
        "tag": tag,
    }

if __name__ == "__main__":
    df = pd.read_parquet(DATA / "table_completed_preprocessed.parquet")

    pipeline_env = os.getenv("PIPELINE_NAME")
    model_env = os.getenv("MODEL_NAME")
    pipelines = [pipeline_env] if pipeline_env else ["cost_sensitive"]
    base_models = [model_env] if model_env else ["xgboost"]

    for pipeline_name in pipelines:
        for model_name in base_models:
            results = run_pipeline(df, model_name, pipeline_name)
            tag = f"{pipeline_name}-{model_name}"

            path_models = RESULTS / "models" / model_name / pipeline_name
            path_alz = RESULTS / "alzheimer" / model_name / pipeline_name
            path_esq = RESULTS / "esquizofrenia" / model_name / pipeline_name

            for path in [path_models, path_alz, path_esq]:
                path.mkdir(parents=True, exist_ok=True)

            # Guardar modelo entrenado y métricas de CV
            with open(path_models / "model.pkl", "wb") as f:
                pickle.dump(results["best_model"], f)

            metrics_cv_df = pd.DataFrame([results["cv_metrics"]])
            metrics_cv_df["pipeline"] = tag
            metrics_cv_df["best_params"] = str(results["params"])
            metrics_cv_df.to_csv(path_models / "metrics.csv", index=False)

            # Guardar resultados Alzheimer
            df_k_alz, summ_alz, report_alz, df_preds_alz = results["alz"]
            df_preds_alz["pipeline"] = tag
            df_preds_alz.to_csv(path_alz / "predictions.csv", index=False)
            df_k_alz["pipeline"] = tag
            df_k_alz.to_csv(path_alz / "recall_precision_at_k.csv", index=False)
            df_metrics_alz = pd.DataFrame([summ_alz])
            df_metrics_alz["pipeline"] = tag
            df_metrics_alz.to_csv(path_alz / "metrics.csv", index=False)

            # Guardar resultados Esquizofrenia
            df_k_esq, summ_esq, report_esq, df_preds_esq = results["esq"]
            df_preds_esq["pipeline"] = tag
            df_preds_esq.to_csv(path_esq / "predictions.csv", index=False)
            df_k_esq["pipeline"] = tag
            df_k_esq.to_csv(path_esq / "recall_precision_at_k.csv", index=False)
            df_metrics_esq = pd.DataFrame([summ_esq])
            df_metrics_esq["pipeline"] = tag
            df_metrics_esq.to_csv(path_esq / "metrics.csv", index=False)

            print(f"\n Finalizado: {model_name} | {pipeline_name}")